%% ERGO COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1993 by The University of Queensland, Queensland 4072 Australia
%% 
%% 
%% Permission to use, copy and distribute this software for any
%% non-commercial purpose and without fee is hereby granted, provided
%% that the above copyright notice and this permission notice and
%% warranty disclaimer appear in all copies and in supporting
%% documentation, and that the name of the University of Queensland not
%% be used in advertising or publicity pertaining to distribution of
%% the software without specific, written prior permission.
%% 
%% Source code modifications are prohibited except where written 
%% agreement has been given in advance by the University of Queensland.
%% 
%% The University of Queensland disclaims all warranties with regard to
%% this software, including all implied warranties of merchantability
%% and fitness.  In no event shall the University of Queensland be
%% liable for any special, indirect or consequential damages or any
%% damages whatsoever resulting from loss of use, data or profits,
%% whether in an action of contract, negligence or other tortious
%% action, arising out of or in connection with the use or performance
%% of this software.
%% 
%% CVS theory_data.ql,v 1.7 1993/12/10 02:23:20 marku Exp


?- op(400, yfx, (/)).
?- op(810, xfy, (and)).
?- op(830, xfy, (=>)).
?- op(840, xfy, (<=>)).



/*
**  primitve functions, primitive quantifiers and obvars
*/


const(Name)        :- primf(Name, default, 0).
primf(Name, Arity) :- primf(Name, default, Arity).
primq(Name, Prec)  :- primf(Name, Prec,    quant).

primf(Name, Prec, OpInfo) :-
	check_open,
	claim_symbol(Prec, OpInfo, Name).

obj_var(Name) :- 
	claim_symbol(obvar, obvar, Name).

/* 
** declare a primitive axiom 
*/

mkaxiom(Name, Term) :-
	mkaxiom(Name, Term, true).

%% For backwards compatability.
mkaxiom(Name, Term, NFI) :- 
	isList(NFI),
	!,
	check_atom(Name),
	available(Name),
	check_open,
	check_term(Term),
	check_nfi(NFI),
	assert_thy(axiom, [Name, Term]).

mkaxiom(Name, Term, Body) :- 
	check_atom(Name),
	available(Name),
	check_open,
	check_term(Term),
	assert_thy_body(axiom, [Name, Term], Body).


/* 
** define a function 
*/

deff(Name, ObVars, Term) :-
	length(ObVars, OpInfo),
	deff(Name, ObVars, Term, default, OpInfo).

deff(Name, ObVars, Term, Prec, OpInfo) :- 
	call((legal_func_term(Term, ObVars),
	LHS =.. [Name|ObVars],
	rep_obs_with_vars(LHS = Term, ObVars, Axiom),
	claim_symbol(Prec, OpInfo, Name),
	assert_thy(axiom, [def(Name), Axiom]))).

/* 
** define a quantifier 
*/

defq(Name, Bound, Body, Definition, Prec) :- 
	var(Body),
	is_object_var(Bound),
	legal_quant_term(Definition, Bound, Body),
	claim_symbol(Prec, quant, Name),
	quantify(Term, Name, Bound, Body),
	assert_thy(axiom, [def(Name), Term = Definition]).


/*
** Add a tactic file
*/


add_tactic(File) :-
	once(get_up(tactic_file, [File])),
        writelist(['The tactic file ', File, '  is already loaded.']),nl,!.

add_tactic(File) :-
	do_consult(File),
	assert_thy(tactic_file, [File]).


/*
** Add a global tactic file
*/

add_global_tactic(File) :-
	(tactic_file : [File]).

add_global_tactic(File) :-
	loaded(File), !,
	writelist(['The tactic file ', File, ' is already loaded.']),nl,
	fail.

add_global_tactic(File) :-
	current_dir(CD), !,
	abolish(current_dir,1), 
	do_consult(File),
	assert_global(tactic_file, [File]),
	assert(current_dir(CD)).

add_global_tactic(File) :-
	do_consult(File),
	assert_global(tactic_file, [File]).

/*
** Note an equivalence partial ordering
*/

is_stronger_than(Op1, Op2) :-
	preorder(Op1),
	preorder(Op2),
	LHS =.. [Op1, A, B],
	RHS =.. [Op2, A, B],
	rule_or_complain(LHS => RHS, Complain),
	var(Complain),
	assert_thy(stronger_than, [Op1, Op2]),
	mk_preorder_info.

/*
** Note than an operator is an equivalence operator
*/

is_equivalence(Equ) :-
	once((
	        preorder(Equ)
	    ;
	        is_preorder(Equ)
	    )),
	ST1 =.. [Equ, B, C],
	ST2 =.. [Equ, C, B],
	rule_or_complain(ST1 => ST2, Complain),
	var(Complain),
	assert_thy(equivalence, [Equ]),
	mk_preorder_info.

is_preorder(Op) :-
        R =.. [Op, A, A],
        rule_or_complain(R, Complain),
        T1 =.. [Op, D, E],
        T2 =.. [Op, E, F],
        T3 =.. [Op, D, F],
        rule_or_complain((T1 and T2) => T3, Complain),
	var(Complain),
        assert_thy(preorder, [Op]),
	mk_preorder_info.

is_reverse_of(Op1, Op2) :-
	Op1 \== Op2,
        ST1 =.. [Op1, B, C],
        ST2 =.. [Op2, C, B],
        rule_or_complain(ST1 <=> ST2, Complain),
	var(Complain),
        assert_thy(reverse_of, [Op2, Op1]),
        assert_thy(reverse_of, [Op1, Op2]),
	mk_preorder_info.

	

rule_or_complain(A, _) :-
	freeze(A),
	is_theorem(_, A), !.

rule_or_complain(A, true) :-
	(writelist(['Missing Obligation: ', A, nl]), fail) ; true.


