%% ERGO COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1993 by The University of Queensland, Queensland 4072 Australia
%% 
%% 
%% Permission to use, copy and distribute this software for any
%% non-commercial purpose and without fee is hereby granted, provided
%% that the above copyright notice and this permission notice and
%% warranty disclaimer appear in all copies and in supporting
%% documentation, and that the name of the University of Queensland not
%% be used in advertising or publicity pertaining to distribution of
%% the software without specific, written prior permission.
%% 
%% Source code modifications are prohibited except where written 
%% agreement has been given in advance by the University of Queensland.
%% 
%% The University of Queensland disclaims all warranties with regard to
%% this software, including all implied warranties of merchantability
%% and fitness.  In no event shall the University of Queensland be
%% liable for any special, indirect or consequential damages or any
%% damages whatsoever resulting from loss of use, data or profits,
%% whether in an action of contract, negligence or other tortious
%% action, arising out of or in connection with the use or performance
%% of this software.
%% 
%% CVS unif.ql,v 1.5 1993/12/10 02:23:39 marku Exp


?- object_var(x).
?- op(600,xfy,nfi).  %% same as not_free_in


/*
 * with_frozen_nfi_test(+Flag, +Code)
 *
 *  Executes Code with frozen_nfi_test flag set to Flag.
 *  It fails iff Code fails.
 *  It always resets the frozen_nfi_test to its original value.
 */

with_frozen_nfi_test(Flag, Code) :-
    setFlag(frozen_nfi_test, Flag, Orig),
    Code,
    setFlag(frozen_nfi_test, Orig, _).


/*
 * A user-level predicate for adding not_free_in conditions.
 */

nfi(Var,Term) :-
    with_frozen_nfi_test(off, Var not_free_in Term).



complete_retry_delays :-
    '$check_delayed_problem'('='), !,
    retry_delay_problems,
    \+ '$check_delayed_problem'('=').

complete_retry_delays.

/*
 * incomplete_retry_delays
 * attempts to solve any unsolved unification problems
 * using incomplete_unify
 */

iu :- incomplete_retry_delays. 

incomplete_retry_delays :-
	retry_delay_problems,
	incomplete_unify_problem.

incomplete_unify_problem :-
	retry_delay_problems,
        (   get_unify_delays([(T1 = T2)|_])
	    ->
		writeln(top(T1 = T2)), flush(user_output),

		setFlag(frozen_nfi_test, on, Orig), /*PJR fudge*/
			incomplete_unify(T1, T2),
		setFlag(frozen_nfi_test, Orig,_), /*PJR fudge*/

		incomplete_unify_problem
	    ;
		true
	).


/*
 * incomplete_unify(T1, T2) attempts to find a unifier
 * of T1 and T2 - not necessarily the MGU.
 * It will typically instantiate bound variables where possible
 * rather than do change of bound variable.
 * Similarly, in matching variables it attempts to make their
 * corresponding substitutions identical.
 * This is a backtracking algorithm that searches for unifiers.
 *
 */




incomplete_unify(T1, T2) :-
        simplify_term(T1, ST1),
        simplify_term(T2, ST2),
        iu_2(ST1, ST2).


iu_2(X,Y) :-
        var(X),
        var(Y),
        iu_var_var(X,Y).
iu_2(X,T) :-
        var(X),
        nonvar(T),
        iu_var(X,T).
iu_2(T,X) :-
        var(X),
        nonvar(T),
        iu_var(X,T).
iu_2(X,Y) :-
        is_object_var(X),
        is_object_var(Y),!,
        iu_obvar_obvar(X,Y).
iu_2(X,T) :-
        is_object_var(X),
        structure(T),
        iu_obvar(X,T).
iu_2(T,X) :-
        is_object_var(X),
        structure(T),
        iu_obvar(X,T).


iu_var_var(X,Y) :-
        break_term(X, SX, BX),
        break_term(Y, SY, BY),
        ((pair_subs(SX,BX, SY,BY, SX1, SY1),
	  bind_terms(X, Y, SX1, BX, SY1, BY))
	;
	  BX = x
	;
	  BY = x).

bind_terms(XT, YT, SX, X, SY, Y) :-
        X == Y,
	simplify_term(XT, SimpXT),
	simplify_term(YT, SimpYT),
	substitute(SimpXT, SimpSX, _),
	substitute(SimpYT, SimpSY, _),
        flatten_sub(SimpSX, FSX),
        flatten_sub(SimpSY, FSY),
	make_progress(XT, YT).

	% FSX == FSY.

bind_terms(_, _, SX, X, SY, Y) :-
	\+  (X == Y),
	SX == [],
	\+  frozen(X),
	!,
        flatten_sub(NSY, SY),
	substitute(NY, NSY, Y),
	X = NY,
	make_progress(X, NY).
bind_terms(_, _, SX, X, SY, Y) :-
	\+  (X == Y),
	SY == [],
	\+  frozen(Y),
	!,
        flatten_sub(NSX, SX),
	substitute(NX, NSX, X),
	NX = Y,
	make_progress(NX, Y).
bind_terms(_, _, SX, X, SY, Y) :-
	\+  (X == Y),
        flatten_sub(NSX, SX),
	substitute(TX, NSX, X),
        flatten_sub(NSY, SY),
	substitute(TY, NSY, Y),
	X = Y,
	TX == TY.

pair_subs(SX, X, [], Y, SX1, SY1) :-
	!,
	pair_subs1([], Y, SX, X, SY1, SX1).
pair_subs(SX,BX, SY,BY, SX1, SY1) :-
	pair_subs1(SX,BX, SY,BY, SX1, SY1).

pair_subs1([],X, SY,Y, [], NSY) :-
	delete_entries(SY,Y, NSY).
pair_subs1([(R/D)|T],X, SY,Y, NSX, NSY) :-
        iu_compatable(R,D,SY,CSY),
	pair_subs1(T,X,CSY,Y, NSX, NSY).
pair_subs1([(D/D)|T],X, SY,Y, NSX, NSY) :-  /* R = D */
        pair_subs1(T,X, SY,Y, NSX, NSY).
pair_subs1([(R/D)|T],X, SY,Y, NSX, NSY) :-
	iu_delete(R,D, SY, SY1),
	pair_subs1(T,X, SY1,Y, NSX, NSY).
pair_subs1([HX|TX],X, SY,Y, [HX|NSX], NSY) :-
        pair_subs1(TX,X, SY,Y, NSX, NSY).
pair_subs1([(R/D)|T],X, SY,Y, NSX, NSY) :-
        \+  frozen(D),
	D not_free_in X,
        pair_subs1(T,X, SY,Y, NSX, NSY).



iu_compatable(R,D, [(RY/DY)|T], T) :-
	D == DY,
	R = RY.
iu_compatable(R,D, [E|T], [E|NT]) :-
        iu_compatable(R,D,T, NT).

iu_delete(R,D, [(RY/DY)|T], T) :-
	\+  (D == DY),
	D = DY,
	R = RY.
iu_delete(R,D, [E|T], [E|NT]) :-
        iu_delete(R,D,T, NT).


delete_entries([],Y, []).
delete_entries([H|T],Y, [H|NT]) :-
        delete_entries(T,Y, NT).
delete_entries([(D/D)|T],Y, NT):-    /* D = R */
        delete_entries(T,Y, NT).
delete_entries([(R/D)|T],Y, NT):-
	\+  frozen(D),
	D not_free_in Y,
        delete_entries(T,Y, NT).

	
iu_var(X, T) :-
	% writeln(first_new_bits(X, T)),
	is_object_var(T),
	simplify_term(T, TS),
	substitute(TS, YS, Y),
	flatten_sub(YS, FYS),
	last_of(FYS, (R/D)),
	% writeln('New Bit'),
	(
		% writeln(new(Y = D)), 
		Y = D
	;
		% writeln('try_nfi_new'), 
		Y not_free_in D
		% , writeln(new(Y, D))
	).
	

iu_var(X, T) :-
	\+  is_object_var(T),
        nonvar(T),
        break_term(X, SX, BX),
        iu_var_2(X, SX,BX,T).

iu_var_2(_,[], T,T).
iu_var_2(_, S1, X, T) :-
        S1 \== [],
        structure(T),
        skel(T, X).
iu_var_2(X,S1, BX, T) :-
        S1 \== [],
        BX =  x.


iu_obvar_obvar(X,Y) :-
	break_term(X, SX, BX),
	break_term(Y, SY, BY),
	iu_obvar_reduce(SX, BX, SY, BY).

iu_obvar_reduce([], X, SY, Y) :-
	last_of(SY, (R/D)),
	(
		% writeln(k(Y, D)),
		Y = D
		% , writeln('kkk')
	    ;
		% writeln((Y,D)),
		Y not_free_in D
		% , writeln(bye)
	).
iu_obvar_reduce(SX,X,SY,Y) :-
	last_of(SX, (R/D)),
	(
		% writeln(p(X = D)),
		X = D
	    ;
		% writeln(p(X, D)),
		X not_free_in D
	).

iu_obvar(X,T) :-
        break_term(X, SX, BX),
	last_of(SX, (R/D)),
	(
	        BX = D
	    ;
	        BX not_free_in D
	).
	



skel(T, SkelT) :-
        T =.. [Op|Args],
        length(Args, AL),
        length(NArgs, AL),
        SkelT =.. [Op|NArgs].

skel(T, SkelT) :-
        quantify(T, Q, X, B),
        quantify(SkelT, Q, x, A).



break_term(T, ST, BT) :-
    simplify_term(T, NT),
    substitute(NT, S1, BT),
    flatten_sub(S1, ST).

flatten_sub([], []) :- !.
flatten_sub([X], X).

structure(T) :-
        \+ var(T),
        \+ is_object_var(T).

make_progress(A,B) :-
	setFlag(frozen_nfi_test, off, Orig),
	retry_delay_problems,
	(get_unify_delays(P) ->
	    setFlag(frozen_nfi_test, Orig,_),
	    make_progress2(A,B,P)
	;
	    setFlag(frozen_nfi_test, Orig,_)
	).

make_progress2(A,B,L) :-
        \+ found_delay_problem(A,B,L).


found_delay_problem(A,B,[H|T]) :-
	H == (A = B) ; H == (B = A).
found_delay_problem(A,B,[_|T]) :-
	found_delay_problem(A,B,T).



last_of([H], H).
last_of([H1,H2|T], Last) :-
	last_of([H2|T], Last).
