%% This 'ergo' theory is a minimal theory that corresponds
%% to the builtin assumptions made by Ergo 4.0.
%%
%% It declares the binary operators, 'and', '=>' and '='
%% and defines '<=>' to be the equivalence relation induced by '=>'.

add_theory(ergo).

constant true.
function 700 nonassoc '='(_,_).
function 810 rightassoc 'and'(_,_).
function 830 rightassoc '=>'(_,_).

% The following declaration is not necessary, but highly recommended,
% so that ':' can be used for typing notation.  The ':' notation for
% typing will be primitive in future versions of Ergo (and Qu-Prolog 4).
primf(:,650,xfx).

axiom true
=== true.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Properties of =>
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

axiom refleximp
=== A => A.

postulate transimp
=== ((A => B) and (B => C)) => (A => C).

is_preorder(=>).

%% The following axiom states that 'true' is the weakest predicate.
%% It is a useful simplification rule, so we put it in the 'standard'
%% rule group, which is used by 'focus' to perform simplifications.
axiom imp_true
=== A => true.

%% The following axiom states that any predicate that has 
%% been proven is *equal* to 'true'.
%% This property is necessary for trans_true in equality.
axiom impeqltrue
===  A => (A = true).

openrule(A => B,  [2], 
	[A], 
	id,
	[=> --> =>],
	[A => G --> G, id]
        ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Minimal properties of 'and'.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

axiom trueandtrue
=== true and true.


openrule(A and B, [1], 
	[B],
	id,
	[=> --> =>],
	id).

openrule(A and B, [2],
	[A],
	id,
	[=> --> =>],
	id).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Properties of =
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

axiom reflexeql 
=== (A = A).

axiom symmeql  
=== ((A = B) => (B = A)).

axiom transeql 
=== (A = B) and (B = C) => (A = C).

is_equivalence(=).

% from this axiom we can also deduce that = is stronger than <=>.
axiom eqlimpimp
=== A = B => (A => B).

is_stronger_than(=, =>).

%% Opening Rules
openrule(A = B, [1],
	 [], 
	 id, 
	 [= --> =],
	 [G = B --> G, true --> B]).

openrule(A = B, [2],
	 [], 
	 id, 
	 [= --> =],
	 [A = G --> G, true --> A]).


% Now that equality is defined, we can start to use rule_groups
% and definitions etc.
rule_group(standard, imp_true, t, _ => true).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% The derived concept: '<=>'
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function 840 rightassoc '<=>'(A, B)
=== (A => B) and (B => A).

%% theorem reflexeqv
%% ===  A <=> A.
%
rerun(ergo:reflexeqv).

%% theorem transeqv
%% ===  ((A <=> B) and (B <=> C)) => (A <=> C).
%
rerun(ergo:transeqv).

%% theorem symmeqv
%% ===  (C <=> A) => (A <=> C).
%
rerun(ergo:symmeqv).


is_equivalence(<=>).


%% theorem eqvimpimp
%% ===  (C <=> A) => (C => A).
%
rerun(ergo:eqvimpimp).

is_stronger_than(<=>, =>).


%% theorem eqlimpeqv
%% ===  (C = A) => (C <=> A).
%
rerun(ergo:eqlimpeqv).

is_stronger_than(=, <=>).


openrule(A <=> B, [1], 
	[],
	id,
	[= --> <=>],
	[C <=> D --> C, true --> B]).

openrule(A <=> B, [2], 
	[],
	id,
	[= --> <=>],
	[C <=> D --> D, true --> A]).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Additional properties of 'and'
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% theorem andtrue
%% === A and true <=> A.
%
rerun(ergo:andtrue).

%% theorem trueand
%% === true and A <=> A.
%
rerun(ergo:trueand).


rule_group(standard, andtrue, f).
rule_group(standard, trueand, f).

%% theorem symand
%% === (A and B) <=> (B and A).
%
rerun(ergo:symand).

%% theorem and_assoc
%% === ((A and B) and C) <=> (A and (B and C)).
%
rerun(ergo:and_assoc).


