%% ERGO COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1993 by The University of Queensland, Queensland 4072 Australia
%% 
%% 
%% Permission to use, copy and distribute this software for any
%% non-commercial purpose and without fee is hereby granted, provided
%% that the above copyright notice and this permission notice and
%% warranty disclaimer appear in all copies and in supporting
%% documentation, and that the name of the University of Queensland not
%% be used in advertising or publicity pertaining to distribution of
%% the software without specific, written prior permission.
%% 
%% Source code modifications are prohibited except where written 
%% agreement has been given in advance by the University of Queensland.
%% 
%% The University of Queensland disclaims all warranties with regard to
%% this software, including all implied warranties of merchantability
%% and fitness.  In no event shall the University of Queensland be
%% liable for any special, indirect or consequential damages or any
%% damages whatsoever resulting from loss of use, data or profits,
%% whether in an action of contract, negligence or other tortious
%% action, arising out of or in connection with the use or performance
%% of this software.
%% 
%% CVS rerun.ql,v 1.11 1993/12/10 02:19:03 marku Exp


:- declare(rerun/1).
:- declare(force/1).

/*
** Rerun(Name, SaveFlag)
**
** rerun the proof of a theorem, maybe saving an updated proof file.
*/



force(Thm) :- rerun(Thm, false).
rerun(Thm) :- rerun(Thm, true).

rerun(Thm, Flag) :-
	once((decode_rule(Thm, CD, Name), 
	      atom(Name),
	      cd(CD),
	      get_proof_record(CD:Name, proof_of(_, Form, Ver, Proof)))),
	theorem(CD:Name, Form, 
		(ipSet(abandon_proof_record, Flag),
		 ver_msg(Ver),
		 safe(rerun_proof(Ver, Proof)))).


/*
**  Save / Don't save / Really Don't Save
*/

save_proof :-
	check_or_writelist(
		\+ ipLookup(abandon_proof_record, very_true),
		['Proof record in unsaveable state\n']),
	ipSet(abandon_proof_record, false).

dont_save_proof :-
	ipLookup(abandon_proof_record, very_true), !.

dont_save_proof :-
	ipSet(abandon_proof_record, true).



/*
**
*/


ver_msg('1.3') :- !.

ver_msg(X) :- 
	member(X, ['1.2', '1.1', 1]), !,
	write('This is an obsolete proof format  '), 
	check_update, !.

ver_msg(B) :-
        writelist(['Version ', B, ' proof records not supported by rerun\n']), 
	!, fail.

check_update :-
	ipLookup(abandon_proof_record, true), 
	question('Update stored proof? ') 
	->
	ipSet(abandon_proof_record, false)
	;
	true.


/*
** Rerun the proof steps
*/

rerun_proof('1.3', (_, FL, Hyps)) :- 
	!, rerun_hyps13(Hyps, List),
	rerun_foci('1.3', FL, List).

rerun_proof(V, (_, FL, Hyps)) :-
	rerun_hyps(V, Hyps),
	rerun_foci(V, FL, Hyps).


rerun_foci(V, [], Hyps).
rerun_foci(V, [(F,Code)|FL], Hyps) :-
        rerun_foci(V, FL, Hyps),
	once(rerun_focus(V,F,Code,Hyps)).

rerun_focus(_, F, initial_focus, _) :-
	is_focus(F), incomplete_retry_delays, !.
rerun_focus('1.3', F, trans_dir(A, B, C), Hyps) :-
	lookup([A|C], Hyps, [A1|C1]),
	trans(A1, B, F, C1), !.
rerun_focus(V, F, trans_dir(A, B, C), Hyps) :-
	V \== '1.3',
	translate_rules([A|C], [A1|C1], Hyps),
	trans(A1, B, F, C1), !.
rerun_focus(_, F, wrap(N, T, Foci), Hyps) :-
        wrap(N, T, once(rerun_foci(V, Foci, Hyps))), !. 
rerun_focus(V, F, wrap(N, T, J, Foci), Hyps) :-
        wrap(N, T, once(rerun_foci(V, [(F,J)|Foci], Hyps))), !. 
rerun_focus(V, F, open_pos(_, Pos, Proof), _) :-
        open_pos(Pos, Pos, [], rerun_proof(V, Proof)),
	is_focus(F), incomplete_retry_delays, !.
rerun_focus(V, F, J, H) :-
	(V == '1.3' -> writeln(H) ; true),
	writelist(['FAIL: ', F, '  by  ', J, '\n']), it(_).



%% This reruns all the hypotheses, starting from the end of the list.
%% It's written in this slightly messy style (which requires handling
%% split_hyp twice) to allow for reasonable error messages for each
%% hypothesis that fails, but going through the list once only
%% in a deterministic fashion.


rerun_hyps(V, []).

rerun_hyps(V, [H1,H2|List]) :-
	H1 = (_,_,_,split_hyp(_)),
	!,
	rerun_hyps(V, List),
	once(rerun_hyp(V, H1, List)).  %% and ignore H2

rerun_hyps(V, [Hyp|List]) :-
	rerun_hyps(V, List),
	once(rerun_hyp(V, Hyp, List)).


rerun_hyp(V, (N, T, _, wrap(A, B, Exp)), List) :-
	rerun_hyp(V, (N, T, _, Exp), List), !.

rerun_hyp(_, (_, T, _, take_theorem(R)), List) :-
	take_theorem(R, T, _).

rerun_hyp(_, (_, T, _, h_trans_dir(R, A, B, C)), List) :-
	translate_rules([R,A|C], [R1,A1|C1], List),
	h_trans(R1, A1, B, T, C1).

rerun_hyp(_, (_, _, _, split_hyp(Nr)), List) :-
	translate_rules([Nr], [Nr1], List),
	split_hyp(Nr1).

rerun_hyp(V, (_, T, _, open_hyp(_, Nr, Proof)), List) :-
	translate_rules([Nr], [Nr1], List),
	open_hyp(Nr1, rerun_proof(V, Proof)),
	is_hyp(_, T), 
        incomplete_retry_delays.

rerun_hyp(_, (_, T, _, inherited), _) :-
	is_hyp(_, T), incomplete_retry_delays.

rerun_hyp(_, (_, T, _, translated(_)), _) :-
        is_hyp(_, T), incomplete_retry_delays.

rerun_hyp(_, (N, T, _, J), _) :-
	writelist(['WARNING: h ', N, ':', T, ' was not ', J, '\n']).


%%
%% Conjecture rerun strategy:
%% 
%% 1) Abandon unjustified and part-justified hypotheses & conjectures
%%    Check all inherited/translated hypotheses
%%    Restate all conjectures
%%    Rerun all remaining hypotheses
%%
%% 5) Discharge all conjectures
%%

                            
rerun_hyps13(HL, Lookup) :-
	check_hyp13(HL, Lookup, Conj),
	rerun_conjectures(Conj).

check_hyp13([], [], []).
check_hyp13([A|B], L2, C2) :-
	check_hyp13(B, L, C),
	check_hyp13(A, L, L2, C, C2).


check_hyp13((_, _, _, _, Dep), L, L, C, C) :-
	'$collect_vars'(Dep, var, [], [_|_]), !.
check_hyp13((N, T, _, IT, _), L, [(N, N1) | L], C, C) :-
	once((IT == inherited ; IT = translated(_) ; IT == introduced)),
	is_hyp(N1, T), !.
check_hyp13((N,T,_,conjecture(_,Just),_),  L, [(N,N1)|L], C, [(N1,Just)|C]) :-
	conjecture(T, N1), !.
check_hyp13((N,T,_,take_theorem(Name),_), L, [(N,N1)|L], C, C) :-
	lookup([Name], L, [Name1]),
	take_theorem(Name1, T, N1), !.
check_hyp13((N,T,_,split_left(Nr), _), L, [(N,N1)|L], C, C) :-
	member((Nr, Nr1), L),
	split_left(Nr1, N1), !.
check_hyp13((N,T,_,split_right(Nr), _), L, [(N,N1)|L], C, C) :-
        member((Nr, Nr1), L),
        split_right(Nr1, N1), !.
check_hyp13((N,T,_,h_trans_dir(Name, Rule, Dir, List), _), L, [(N,N1)|L], C, C) :-
	lookup([Name, Rule | List], L, [Name1, Rule1 | List1]),
	h_trans(Name1, Rule1, Dir, T, List1, N1), !.
check_hyp13((N,T,_,open_hyp(_, Nr, Proof), _), L, [(N,N1)|L], C, C) :-
	member((Nr, Nr1), L),
	open_hyp(Nr1, rerun_proof('1.3', Proof), N1), !.
check_hyp13((N, T, _, J, _), L, L, C, C) :-
	writelist(['WARNING: h ', N, ':', T, ' was not ', J, '\n']).


lookup([], _, []).
lookup([A|B], L, [C|D]) :-
	integer(A), !,
	member((A, C), L), !,
	lookup(B, L, D).
lookup([A|B], L, [A|D]) :-
	lookup(B, L, D).


rerun_conjectures([]).
rerun_conjectures([(N, Proof) | Rest]) :-
	discharge(N, rerun_proof('1.3', Proof)), !,
	rerun_conjectures(Rest).
rerun_conjectures([(N, J) | Rest]) :-
	writelist(['WARNING: conjecture ', N, ' was not ', J, '\n']), !,
	rerun_conjectures(Rest).

	
	
/*
** unix operations
*/


get_proof_record(N, Clause) :-
	proof_file_name(N, File),
	(path : [P]),
	full_file_name(P, File, _, PSFile),
	consult_pf_file(PSFile, Clause), !.


consult_pf_file(FileName, Clause) :-
    	flush(user_error),
    	seeing(CurrentInput),
    	debugpause((
		see(FileName), 
    		(in_proof -> read(C) ; readR(C)),
    		seen
		)), !,
    	see(CurrentInput),
    	C = (Clause :- Body),
    	once(Body).

/*
** Use the proof record hypothesis list to dereference 
** proof record hypothesis numbers
*/


translate_hyp(A : X,          Y) :- translate_hyp(X, Y).
translate_hyp(h(N),          (N, _, _, _)) :- integer(N).
translate_hyp(h(T),          (_, T, _, _)).
translate_hyp((A, B, C, D), (A, B, C, D)).
translate_hyp(N,            (N, _, _, _)) :- integer(N).


translate_rules([], [], Hyps).

translate_rules([A|L], [h(T)|L1], Hyps) :-
	translate_hyp(A, (N, T, _, _)),
	once(member((N, T, _, _), Hyps)),
	translate_rules(L, L1, Hyps).

translate_rules([A0|L], [A|L], Hyps) :-
	translate_rulename(A0,A),
	!,
	writelist(['Translated rulename ', A0, '  to  ', A, '\n']),
	translate_rules(L, L1, Hyps).

translate_rules([A|L], [A|L1], Hyps) :-
	translate_rules(L, L1, Hyps).

/*
** translate_rulename(+OldName, -NewName).
**
**   There should really be some way of allowing the user to specify
**   this table.
*/
translate_rulename(arith:eval_int(_),      arith:eval_int).
translate_rulename(arith:eval_eq(_,_),     arith:eval_eq).
translate_rulename(arith:eval_lt(_,_),     arith:eval_lt).
translate_rulename(arith:eval_le(_,_),     arith:eval_le).
translate_rulename(arith:eval_not_eq(_,_), arith:eval_not_eq).

translate_rulename(eval_int(_),            arith:eval_int).
translate_rulename(eval_eq(_,_),           arith:eval_eq).
translate_rulename(eval_lt(_,_),           arith:eval_lt).
translate_rulename(eval_le(_,_),           arith:eval_le).
translate_rulename(eval_not_eq(_,_),       arith:eval_not_eq).
