%% ERGO COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1993 by The University of Queensland, Queensland 4072 Australia
%% 
%% 
%% Permission to use, copy and distribute this software for any
%% non-commercial purpose and without fee is hereby granted, provided
%% that the above copyright notice and this permission notice and
%% warranty disclaimer appear in all copies and in supporting
%% documentation, and that the name of the University of Queensland not
%% be used in advertising or publicity pertaining to distribution of
%% the software without specific, written prior permission.
%% 
%% Source code modifications are prohibited except where written 
%% agreement has been given in advance by the University of Queensland.
%% 
%% The University of Queensland disclaims all warranties with regard to
%% this software, including all implied warranties of merchantability
%% and fitness.  In no event shall the University of Queensland be
%% liable for any special, indirect or consequential damages or any
%% damages whatsoever resulting from loss of use, data or profits,
%% whether in an action of contract, negligence or other tortious
%% action, arising out of or in connection with the use or performance
%% of this software.
%% 
%% CVS rt.ql,v 1.13 1993/12/10 02:19:06 marku Exp


?- op(810,xfy,(and)).
?- op(830,xfy,(=>)).
?- op(840,xfy,(<=>)).

/*
** use etc are not such good predictors of trans nowadays,
** as they don't pay any attention to the dependancy information.
*/


use_goal(Th, Dir, G) :-
        is_focus(F),
	is_reln(E),
        thawing_new_vars(
                once((get_th_or_hyp(Th, N2, Form, Body, _),
        	      check_rewrite(_, E, Form, F, G, Dir, Ant),
		      call(Body),
                      incomplete_retry_delays,
		      trans(N2, Dir, G, Ant)))
             ).


use_goal(Th) :-
	use_goal(Th, _).

use_goal(T, G) :-
	is_goal(Goal),
	init_var(G, Goal, G2),
	use_goal(T, _, G2).

use_rule(T) :- use_goal(T, _, _).


%% satisfy(+Conjuncts, ?RuleList, ?ConjectureList)
%%
%%  Attempts to produce (in RuleList) a list of rule names and hypothesis
%%  and conjecture numbers suitable for use in a trans command.
%%  Unsatisfied conjuncts cause a term of the form c(_) to be inserted
%%  into ConjectureList, so that trans will generate conjectures for
%%  those conjuncts.  Thus, calling satisfy with ConjectureList = []
%%  means that satisfy will fail unless all conjuncts are satisfied.

satisfy(C, L) :- satisfy(C, L, []).

satisfy(X, [], []) :-
	X == true,
	!.

satisfy(X, L, CL) :-
	ipLookup(hyplist, HL),
	once((satisfy2(X, HL, L, CL), incomplete_retry_delays)).


satisfy2(X, _, _, _) :-
	var(X),
	X = true,
	!, fail.


satisfy2(X and Y, HL, [Xrule|Yrules], CS) :-
        !, 
	((var(X) -> frozen(X)) ; true),
	satisfy2_rule(X, HL, Xrule, CS, CS2),
	once(satisfy2(Y, HL, Yrules, CS2)). 

satisfy2(X, HL, [Xrule], CS) :-
	satisfy2_rule(X, HL, Xrule, CS, []).

satisfy2_rule(X, HL, T, CS, CS) :-
	member((T, X, D, _), HL),
	var(D).

satisfy2_rule(X, _, T, CS, CS) :-
	once(is_theorem2(T, X)).

satisfy2_rule(X, _, c(N), [c(N)|CS], CS).

	

/*
** find a rewrite that would work (without actually trying it).  
*
** doesn't check for reverse_true transformations.
*/

check_rewrite(Th, _,     F,              F, true, t).
check_rewrite(Th, _,     F = G,          F, G,    f).
check_rewrite(Th, _,     G = F,          F, G,    r).
check_rewrite(Th, (<=>), F <=> G,        F, G,    f).
check_rewrite(Th, (<=>), G <=> F,        F, G,    r).


check_rewrite(Th, Equ,   XF,      F, G,    Dir, Xlist) :-
	nonvar(XF), XF = (X => Form),
        once(check_rewrite(Th, Equ, Form, F, G, Dir)),
        mk_ant_list(X, Xlist).
check_rewrite(Th, Equ,   XF,      F, G,    Dir, []) :-
	once(check_rewrite(Th, Equ, XF, F, G, Dir)).



mk_ant_list(A, [_|CL]) :-
	nonvar(A), A = (B and C), !,
	mk_ant_list(C, CL).
mk_ant_list(A, [_]).


check_rewrite(Th, Form, F, G, trans(Th, Dir, G, Ant)) :-
        is_reln(Equ),
        get_th_or_hyp(Th, _, Form, Body, _),
        check_rewrite(Th, Equ, Form, F, G, Dir, Ant),
	call(Body).


check_rewrite(Th, F, G, Code) :-
	check_rewrite(Th, _, F, G, Code).


check_rewrite(Th, F, G) :-
	check_rewrite(Th, _, F, G, _).

check_rewrite(Th) :-
	is_focus(F),
	is_goal(G),
	check_rewrite(Th, _, F, G, _).


/*
** is_theorem2 - Exactly as is_theorem, but also looks for compiled rules
*/


is_theorem2(Rule, Form) :-
	decode_rule(Rule, Thy, Name),
	look_for_theorem(Thy:Name, Form),
	(
	    rule_at2(Name, Form, Thy)
	;
	    eval_compiled_rule(Thy, Name, _, _, Form, _)
	).


/*
** term_theory(Term, Theories) - Finds all theories inherited by the
**		current theory which inherit all of the function and 
**		quantifier symbols in the specified term
**
**	NOTE -- does not check the contents of any substitutions for
**		function/quant symbols  (cf check_term/1 in syntax1.ql)
*/


term_theory(Term, Theories) :-
	current_dir(T),
	ancestor : [T, L],
	term_theory(Term, [T | L], Theories).
	
term_theory(Term, [], []).
term_theory(Term, In, In) :-
	var(Term), !.
term_theory(Term, In, Out) :-
	quantify(Term, Q, _, B), !,
	inherited_symbol(Q, quant, Theories),
        intersect(In, Theories, Mid),
        term_theory(B, Mid, Out).
term_theory(Term, In, Out) :-
    	Term =.. [O|Args], !,
    	length(Args, Arity),
    	inherited_symbol(O, Arity, Theories),
    	intersect(In, Theories, Mid),
    	term_list_theory(Args, Mid, Out).

term_list_theory([], In, In).
term_list_theory([A | B], In, Out) :-
	term_theory(A, In, Mid),
	term_list_theory(B, Mid, Out).
	

	
