%% ERGO COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1993 by The University of Queensland, Queensland 4072 Australia
%% 
%% 
%% Permission to use, copy and distribute this software for any
%% non-commercial purpose and without fee is hereby granted, provided
%% that the above copyright notice and this permission notice and
%% warranty disclaimer appear in all copies and in supporting
%% documentation, and that the name of the University of Queensland not
%% be used in advertising or publicity pertaining to distribution of
%% the software without specific, written prior permission.
%% 
%% Source code modifications are prohibited except where written 
%% agreement has been given in advance by the University of Queensland.
%% 
%% The University of Queensland disclaims all warranties with regard to
%% this software, including all implied warranties of merchantability
%% and fitness.  In no event shall the University of Queensland be
%% liable for any special, indirect or consequential damages or any
%% damages whatsoever resulting from loss of use, data or profits,
%% whether in an action of contract, negligence or other tortious
%% action, arising out of or in connection with the use or performance
%% of this software.
%% 
%% CVS tools.ql,v 1.12 1993/12/10 02:19:13 marku Exp


?- op(999, fx, h).
:- declare(db/1).
:- declare(db/0).


/*
** Null tactics
*/

not_used.
nil.


/*
** Display a tactic and ask for confirmation
*/


confirm(X) :- 
	nl, nl, pretty(X),
        question('Proceed').


write_elems([A|B]) :- !, writelist([A|B]).
write_elems(A) :- writeR(A, _).




%% msg(+MessageList)
%%
%%  This is the recommended printing mechanism for all tactics.
%%
%%  It prints MessageList if the current verbosity number is greater 
%%  than zero (see set_verbosity/1 and less_verbose/2).
%%  MessageList is automatically indented and followed by a
%%  newline, except when it is the empty list, in which case 
%%  nothing is printed. 

msg(Msg) :-
    ipLookup(verbosity, V),
    integer(V),
    ipLookup(cw, CW),
    integer(CW),
    !,
    ((((V =< 0) ; (Msg == [])) ->
	true)
    ;
	(writelist([tab(CW)]),
	 writelist_indent(Msg,[tab(CW)]),
	 nl)
    ).

% outside a proof, or verbosity not set.
msg(Msg) :-
ipLookup(verbosity, V),
ipLookup(cw, CW),
writelist(['WARNING: verbosity = ',V, ' and window level = ',CW,nl]),
    ((Msg == [] ->
	true)
    ;
	(writelist(Msg),
	 nl)
    ).


%% set_verbosity(+N).
%%
%%  Determines how many messages will be displayed by tactics.
%%  Zero means no messages.  Higher numbers mean more messages.

set_verbosity(N) :-
    N >= 0,
    ipSet(verbosity,N).


%% less_verbose(+Tactic).
%% less_verbose(+N, +Tactic).
%%
%%  This subtracts N (which defaults to 1 in the first form) 
%%  from the current verbosity level during the execution of Tactic.
%%  NB. Negative values of N *raise* the verbosity level for Tactic.

less_verbose(Tac) :-
    less_verbose(1, Tac).

less_verbose(N, Tac) :-
    ipLookup(verbosity,V0),
    V is V0 - N,
    ipSet(verbosity, V),
    ctac(Tac),
    ipSet(verbosity, V0).

	

/*
**  Display window with additional information
*/


trial_reportIt :- reportIt.
trial_reportWindow :- reportWindow.

reportIt :-
        once(reportWindow), !.

reportIt.


reportWindow :-
    is_focus(Focus),
    ipLookup(cw,       CW),
    ipLookup(goal,     Goal),
    ipLookup(relation,      Equ),
    ipLookup(newvars,  All),
    nl, list_hyps_rev,
    (
	(All == []    %% '$collect_variables'(All, meta_var, [], [A|B]) 
         -> 
	 true)
     ;
	 writelist([nl,'    new vars: ',All, nl])
    ),
    writelist([nl, tab, 'focus:  ', Focus]),
    writelist([nl, tab, Equ]),
    writelist([nl, tab, 'goal:   ', Goal, nl]).


list_hyps_rev :-
    ipLookup(hyplist, Hyps),
    ipLookup(focusdep, FD),
    list_hyps_rev(FD, Hyps).

list_hyps_rev(FD, [(No, Hyp, D, J, Dep)|Rest]) :-
    var(D), !,
    list_hyps_rev(FD, Rest),
    write_hyp(No, Hyp, J, Dep, FD).

list_hyps_rev(FD, [_|Rest]) :-
    !, list_hyps_rev(FD, Rest).

list_hyps_rev(_, _).

write_hyp(No, Hyp, conjecture(_,_), Dep, FD) :-
    var(Dep), 
    \+ \+ (thaw(Dep), Dep = FD), !,
    writelist(['    ', No, '? ', Hyp, nl]).

write_hyp(No, Hyp, conjecture(_,_), Dep, FD) :-
    var(Dep), !,
    writelist(['    ', No, '* ', Hyp, nl]).

write_hyp(No, Hyp, _, Dep, _) :-
    '$collect_variables'(Dep, var, [], []), !,
    writelist(['    ', No, ': ', Hyp, nl]).

write_hyp(No, Hyp, _, Dep, _) :-
    writelist(['    ', No, '% ', Hyp, nl]).

/*
** Type check
*/


hyp_form(h _) :- !.
hyp_form(newest).
hyp_form(X) :- integer(X).





/*
** Simple to use theorem grabber.
*/

th_or_hyp(T, Pattern) :-
	th_or_hyp(T, Pattern, _).

th_or_hyp(Name, Pattern, Nr) :-
	is_hyp(Name, Nr, Pattern).

th_or_hyp(Th, Pattern, Th) :-
        is_theorem(Th, Pattern),
	incomplete_retry_delays.



/*
**
*/

force_tty(Code) :-
	macro_stack([Top|_]), !,
	see(user_input),
	call_tac(Code),
	see(Top).

force_tty(Code) :-
	call_tac(Code).



db(T) :-
	spy(T),
	leash(all).

db :- 
	leash(all), 
	trace.


